﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media.Imaging;
using Microsoft.WebMatrix.Extensibility;

namespace Snippets
{
    /// <summary>
    /// A WebMatrix extension that inserts code snippets.
    /// </summary>
    [Export(typeof(IExtension))]
    public class Snippets : ExtensionBase
    {
        /// <summary>
        /// Extension name.
        /// </summary>
        private const string SnippetsString = "Snippets";

        /// <summary>
        /// Stores a reference to the small Snippets image.
        /// </summary>
        private readonly BitmapImage _snippetsImageSmall = new BitmapImage(new Uri("pack://application:,,,/Snippets;component/Snippets-16x16.png", UriKind.Absolute));

        /// <summary>
        /// Stores a reference to the large Snippets image.
        /// </summary>
        private readonly BitmapImage _snippetsImageLarge = new BitmapImage(new Uri("pack://application:,,,/Snippets;component/Snippets-32x32.png", UriKind.Absolute));

        /// <summary>
        /// Stores a Dictionary of snippet name->text.
        /// </summary>
        private Dictionary<string, string> _snippets = new Dictionary<string, string>();

        /// <summary>
        /// Stores a reference to the extension's Ribbon group.
        /// </summary>
        private IRibbonGroup _ribbonGroup;

        /// <summary>
        /// Stores a reference to the editor's "populate EditorContainer" command.
        /// </summary>
        private ICommand _editorContainerCommand;

        /// <summary>
        /// Initializes a new instance of the Snippets class.
        /// </summary>
        public Snippets()
            : base(SnippetsString, "1.0")
        {
            // Get path to snippets directory
            var snippetsPath = Path.Combine(
                Environment.GetFolderPath(System.Environment.SpecialFolder.MyDocuments),
                SnippetsString);

            // Create help message
            var snippetHelp = string.Format(
                CultureInfo.CurrentCulture,
                "Create a \"{0}\" folder in your Documents directory (\"{1}\") and create a separate text file (ex: Sample.txt) inside for each snippet. The file's name identifies the snippet and its contents are added to the current editor document when the snippet is invoked.",
                SnippetsString,
                snippetsPath);

            // Create dictionary of snippet name->text for each .txt file in the snippets directory
            try
            {
                foreach (var snippetsFile in Directory.EnumerateFiles(snippetsPath, "*.txt"))
                {
                    try
                    {
                        _snippets.Add(
                            Path.GetFileNameWithoutExtension(snippetsFile),
                            File.ReadAllText(snippetsFile));
                    }
                    catch
                    {
                        // Ignore file read errors
                    }
                }
            }
            catch
            {
                // Ignore directory enumeration errors
            }

            // Provide a helpful message for a missing/empty snippets directory
            if (!_snippets.Any())
            {
                _snippets.Add(
                    "[How to use Snippets...]",
                    snippetHelp);
            }

            // Create buttons for each snippet
            var snippetsButtons = _snippets
                .Select(s =>
                    new RibbonButton(
                        s.Key,
                        (s.Value == snippetHelp) ?
                            new DelegateCommand(
                                (parameter) =>
                                {
                                    MessageBox.Show(snippetHelp, SnippetsString);
                                }) :
                            new DelegateCommand(
                                (parameter) => GetEditorExt() != null,
                                (parameter) =>
                                {
                                    // parameter is (incorrectly) null, so add this indirection for now
                                    HandleSnippetInvoke(s.Value);
                                }),
                        s.Value,
                        _snippetsImageSmall,
                        _snippetsImageLarge));

            // Add a group/menu button to the Ribbon
            _ribbonGroup =
                new RibbonGroup(
                    SnippetsString,
                    new IRibbonItem[]
                    {
                        new RibbonMenuButton(
                            "Insert",
                            snippetsButtons,
                            _snippetsImageSmall,
                            _snippetsImageLarge),
                    });
            RibbonItemsCollection.Add(_ribbonGroup);
        }

        /// <summary>
        /// Called when a snippet button is invoked.
        /// </summary>
        /// <param name="parameter">Snippet text.</param>
        private void HandleSnippetInvoke(object parameter)
        {
            var snippet = (string)parameter;
            ReplaceSelectedEditorText(snippet);
        }

        /// <summary>
        /// Called when the WebMatrixHost property changes.
        /// </summary>
        /// <param name="oldValue">Old value.</param>
        /// <param name="newValue">New value.</param>
        protected override void OnWebMatrixHostChanged(IWebMatrixHost oldValue, IWebMatrixHost newValue)
        {
            // Clear old values
            if (null != oldValue)
            {
                oldValue.WorkspaceChanged -= new EventHandler<WorkspaceChangedEventArgs>(WebMatrixHost_WorkspaceChanged);
                _editorContainerCommand = null;
            }

            base.OnWebMatrixHostChanged(oldValue, newValue);

            // Get new values
            if (null != newValue)
            {
                newValue.WorkspaceChanged += new EventHandler<WorkspaceChangedEventArgs>(WebMatrixHost_WorkspaceChanged);
                _editorContainerCommand = WebMatrixHost.HostCommands.GetCommand(new Guid("27a0f541-c86c-4f0b-b436-0b50bf9f7ef8"), 10);

                // Set initial visibility
                _ribbonGroup.IsVisible = newValue.Workspace is IEditorWorkspace;
            }
        }

        /// <summary>
        /// Called when the WebMatrixHost's WorkspaceChanged event fires.
        /// </summary>
        /// <param name="sender">Event source.</param>
        /// <param name="e">Event arguments.</param>
        private void WebMatrixHost_WorkspaceChanged(object sender, WorkspaceChangedEventArgs e)
        {
            // Ribbon button should be visible iff the editor workspace is
            _ribbonGroup.IsVisible = e.NewWorkspace is IEditorWorkspace;
        }

        /// <summary>
        /// Replaces the selected editor text.
        /// </summary>
        /// <param name="insertText">Text to replace with.</param>
        private void ReplaceSelectedEditorText(string insertText)
        {
            // Paste the snippet into the editor
            var paste = WebMatrixHost.HostCommands.Paste;
            if (paste.CanExecute(insertText))
            {
                paste.Execute(insertText);
            }
        }

        /// <summary>
        /// Gets an IEditorExt instance if the editor is in use.
        /// </summary>
        /// <returns>IEditorExt reference.</returns>
        private IEditorExt GetEditorExt()
        {
            var editorContainer = new EditorContainer();
            if (_editorContainerCommand != null)
            {
                _editorContainerCommand.Execute(editorContainer);
            }

            return editorContainer.Editor;
        }

        /// <summary>
        /// Concrete implementation of the IEditorContainer interface.
        /// </summary>
        private class EditorContainer : IEditorContainer
        {
            /// <summary>
            /// Gets or sets the IEditorExt.
            /// </summary>
            public IEditorExt Editor { get; set; }
        }
    }
}
